@@ -48,7 +48,7 @@ bool keyboard_modifier_update(struct wlr_keyboard *keyboard) {
 		XKB_STATE_MODS_LATCHED);
 	xkb_mod_mask_t locked = xkb_state_serialize_mods(keyboard->xkb_state,
 		XKB_STATE_MODS_LOCKED);
-	xkb_mod_mask_t group = xkb_state_serialize_layout(keyboard->xkb_state,
+	xkb_layout_index_t group = xkb_state_serialize_layout(keyboard->xkb_state,
 		XKB_STATE_LAYOUT_EFFECTIVE);
 	if (depressed == keyboard->modifiers.depressed &&
 			latched == keyboard->modifiers.latched &&
@@ -121,20 +121,34 @@ void wlr_keyboard_notify_key(struct wlr_keyboard *keyboard,
 
 void wlr_keyboard_init(struct wlr_keyboard *kb,
 		const struct wlr_keyboard_impl *impl, const char *name) {
-	memset(kb, 0, sizeof(*kb));
+	*kb = (struct wlr_keyboard){
+		.impl = impl,
+		.keymap_fd = -1,
+
+		// Sane defaults
+		.repeat_info.rate = 25,
+		.repeat_info.delay = 600,
+	};
 	wlr_input_device_init(&kb->base, WLR_INPUT_DEVICE_KEYBOARD, name);
 
-	kb->impl = impl;
 	wl_signal_init(&kb->events.key);
 	wl_signal_init(&kb->events.modifiers);
 	wl_signal_init(&kb->events.keymap);
 	wl_signal_init(&kb->events.repeat_info);
+}
 
+static void keyboard_unset_keymap(struct wlr_keyboard *kb) {
+	xkb_keymap_unref(kb->keymap);
+	kb->keymap = NULL;
+	xkb_state_unref(kb->xkb_state);
+	kb->xkb_state = NULL;
+	free(kb->keymap_string);
+	kb->keymap_string = NULL;
+	kb->keymap_size = 0;
+	if (kb->keymap_fd >= 0) {
+		close(kb->keymap_fd);
+	}
 	kb->keymap_fd = -1;
-
-	// Sane defaults
-	kb->repeat_info.rate = 25;
-	kb->repeat_info.delay = 600;
 }
 
 void wlr_keyboard_finish(struct wlr_keyboard *kb) {
@@ -153,33 +167,65 @@ void wlr_keyboard_finish(struct wlr_keyboard *kb) {
 
 	wlr_input_device_finish(&kb->base);
 
-	/* Finish xkbcommon resources */
-	xkb_state_unref(kb->xkb_state);
-	xkb_keymap_unref(kb->keymap);
-	free(kb->keymap_string);
-	if (kb->keymap_fd >= 0) {
-		close(kb->keymap_fd);
-	}
+	keyboard_unset_keymap(kb);
 }
 
 void wlr_keyboard_led_update(struct wlr_keyboard *kb, uint32_t leds) {
+	if (kb->leds == leds) {
+		return;
+	}
+
+	kb->leds = leds;
+
 	if (kb->impl && kb->impl->led_update) {
 		kb->impl->led_update(kb, leds);
 	}
 }
 
-bool wlr_keyboard_set_keymap(struct wlr_keyboard *kb,
-		struct xkb_keymap *keymap) {
-	xkb_keymap_unref(kb->keymap);
-	kb->keymap = xkb_keymap_ref(keymap);
-
-	xkb_state_unref(kb->xkb_state);
-	kb->xkb_state = xkb_state_new(kb->keymap);
-	if (kb->xkb_state == NULL) {
-		wlr_log(WLR_ERROR, "Failed to create XKB state");
-		goto err;
+bool wlr_keyboard_set_keymap(struct wlr_keyboard *kb, struct xkb_keymap *keymap) {
+	if (keymap == NULL) {
+		keyboard_unset_keymap(kb);
+		wl_signal_emit_mutable(&kb->events.keymap, kb);
+		return true;
 	}
 
+	struct xkb_state *xkb_state = xkb_state_new(keymap);
+	if (xkb_state == NULL) {
+		wlr_log(WLR_ERROR, "Failed to create XKB state");
+		return false;
+	}
+
+	char *keymap_str = xkb_keymap_get_as_string(keymap, XKB_KEYMAP_FORMAT_TEXT_V1);
+	if (keymap_str == NULL) {
+		wlr_log(WLR_ERROR, "Failed to get string version of keymap");
+		goto error_xkb_state;
+	}
+	size_t keymap_size = strlen(keymap_str) + 1;
+
+	int rw_fd = -1, ro_fd = -1;
+	if (!allocate_shm_file_pair(keymap_size, &rw_fd, &ro_fd)) {
+		wlr_log(WLR_ERROR, "Failed to allocate shm file for keymap");
+		goto error_keymap_str;
+	}
+
+	void *dst = mmap(NULL, keymap_size, PROT_READ | PROT_WRITE, MAP_SHARED, rw_fd, 0);
+	close(rw_fd);
+	if (dst == MAP_FAILED) {
+		wlr_log_errno(WLR_ERROR, "mmap failed");
+		close(ro_fd);
+		goto error_keymap_str;
+	}
+
+	memcpy(dst, keymap_str, keymap_size);
+	munmap(dst, keymap_size);
+
+	keyboard_unset_keymap(kb);
+	kb->keymap = xkb_keymap_ref(keymap);
+	kb->xkb_state = xkb_state;
+	kb->keymap_string = keymap_str;
+	kb->keymap_size = keymap_size;
+	kb->keymap_fd = ro_fd;
+
 	const char *led_names[WLR_LED_COUNT] = {
 		XKB_LED_NAME_NUM,
 		XKB_LED_NAME_CAPS,
@@ -204,40 +250,6 @@ bool wlr_keyboard_set_keymap(struct wlr_keyboard *kb,
 		kb->mod_indexes[i] = xkb_map_mod_get_index(kb->keymap, mod_names[i]);
 	}
 
-	char *tmp_keymap_string = xkb_keymap_get_as_string(kb->keymap,
-		XKB_KEYMAP_FORMAT_TEXT_V1);
-	if (tmp_keymap_string == NULL) {
-		wlr_log(WLR_ERROR, "Failed to get string version of keymap");
-		goto err;
-	}
-	free(kb->keymap_string);
-	kb->keymap_string = tmp_keymap_string;
-	kb->keymap_size = strlen(kb->keymap_string) + 1;
-
-	int rw_fd = -1, ro_fd = -1;
-	if (!allocate_shm_file_pair(kb->keymap_size, &rw_fd, &ro_fd)) {
-		wlr_log(WLR_ERROR, "Failed to allocate shm file for keymap");
-		goto err;
-	}
-
-	void *dst = mmap(NULL, kb->keymap_size, PROT_READ | PROT_WRITE,
-		MAP_SHARED, rw_fd, 0);
-	if (dst == MAP_FAILED) {
-		wlr_log_errno(WLR_ERROR, "mmap failed");
-		close(rw_fd);
-		close(ro_fd);
-		goto err;
-	}
-
-	memcpy(dst, kb->keymap_string, kb->keymap_size);
-	munmap(dst, kb->keymap_size);
-	close(rw_fd);
-
-	if (kb->keymap_fd >= 0) {
-		close(kb->keymap_fd);
-	}
-	kb->keymap_fd = ro_fd;
-
 	for (size_t i = 0; i < kb->num_keycodes; ++i) {
 		xkb_keycode_t keycode = kb->keycodes[i] + 8;
 		xkb_state_update_key(kb->xkb_state, keycode, XKB_KEY_DOWN);
@@ -246,15 +258,13 @@ bool wlr_keyboard_set_keymap(struct wlr_keyboard *kb,
 	keyboard_modifier_update(kb);
 
 	wl_signal_emit_mutable(&kb->events.keymap, kb);
+
 	return true;
 
-err:
-	xkb_state_unref(kb->xkb_state);
-	kb->xkb_state = NULL;
-	xkb_keymap_unref(keymap);
-	kb->keymap = NULL;
-	free(kb->keymap_string);
-	kb->keymap_string = NULL;
+error_keymap_str:
+	free(keymap_str);
+error_xkb_state:
+	xkb_state_unref(xkb_state);
 	return false;
 }
 
