@@ -1,4 +1,3 @@
-#include <limits.h>
 #include <math.h>
 #include <stdbool.h>
 #include <stdlib.h>
@@ -14,11 +13,22 @@ void wlr_box_closest_point(const struct wlr_box *box, double x, double y,
 		return;
 	}
 
+	// Note: the width and height of the box are exclusive; that is,
+	// for a 100x100 box at (0,0), the point (99.9,99.9) is inside it
+	// while the point (100,100) is outside it.
+	//
+	// In order to be consistent with e.g. wlr_box_contains_point(),
+	// this function returns a point inside the bottom and right edges
+	// of the box by at least 1/65536 of a unit (pixel). 1/65536 is
+	// small enough to avoid a "dead zone" with high-resolution mice
+	// but large enough to avoid rounding to zero (due to loss of
+	// significant digits) in simple floating-point calculations.
+
 	// find the closest x point
 	if (x < box->x) {
 		*dest_x = box->x;
-	} else if (x >= box->x + box->width) {
-		*dest_x = box->x + box->width - 1;
+	} else if (x > box->x + box->width - 1/65536.0) {
+		*dest_x = box->x + box->width - 1/65536.0;
 	} else {
 		*dest_x = x;
 	}
@@ -26,8 +36,8 @@ void wlr_box_closest_point(const struct wlr_box *box, double x, double y,
 	// find closest y point
 	if (y < box->y) {
 		*dest_y = box->y;
-	} else if (y >= box->y + box->height) {
-		*dest_y = box->y + box->height - 1;
+	} else if (y > box->y + box->height - 1/65536.0) {
+		*dest_y = box->y + box->height - 1/65536.0;
 	} else {
 		*dest_y = y;
 	}
